# tree.tcl --
#
#	Implementation of a tree data structure for Tcl.
#
# The API is from the tcllib struct::tree 1.16, which is
# Copyright (c) 1998-2000 by Ajuba Solutions.
#

# this package is intended to be a replacement for the
# tcllib struct::tree package.
# This version of the package builds the trees using
# the Tree package by Evan Rempel http://web.uvic.ca/~erempel/tcl
#
# Each node in a tree is named (::Tree::name) with the struct::tree
# index number. Altering the names/locations of nodes directly
# will violate this API (corrupt the tree).

package require Tcl 8.2
package require Tree 2.0

namespace eval ::struct {}

namespace eval ::struct::tree {

    # the tree "names" are used to record the index of each child

    # counter is used to give a unique name for unnamed trees
    variable counter 0

    # commands is the list of subcommands recognized by the tree
    variable commands [list \
            "append"            \
	    "children"		\
	    "cut"		\
	    "destroy"		\
	    "delete"		\
	    "depth"		\
	    "exists"		\
	    "get"		\
            "getall"            \
	    "index"		\
	    "insert"		\
	    "isleaf"		\
            "keys"              \
            "keyexists"         \
            "lappend"           \
	    "move"		\
	    "next"		\
	    "numchildren"	\
	    "parent"		\
	    "previous"		\
	    "set"		\
	    "size"		\
	    "splice"		\
	    "swap"		\
	    "unset"		\
	    "walk"		\
	    ]

    # Only export one command, the one used to instantiate a new tree
    namespace export tree
}

# ::struct::tree::tree --
#
#	Create a new tree with a given name; if no name is given, use
#	treeX, where X is a number.
#
# Arguments:
#	name	Optional name of the tree; if null or not given, generate one.
#
# Results:
#	name	Name of the tree created

proc ::struct::tree::tree {{name ""}} {
    variable counter
    
    if { [llength [info level 0]] == 1 } {
	incr counter
	set name "tree${counter}"
    }

    if { [llength [info commands ::$name]] } {
	error "command \"$name\" already exists, unable to create tree"
    }

    # Set up the namespace
    namespace eval ::struct::tree::tree$name {
	# Set up root node's child list
	variable root
        set root [::Tree::create root]
        ::Tree::attribute $root data ""
        ::Tree::attribute $root _childList {}

        # array to map ::struct::tree names to ::Tree trees
        variable names2nodes
        set names2nodes(root) $root

	# Set up a value for use in creating unique node names
	variable nextUnusedNode
	set nextUnusedNode 1
    }

    # Create the command to manipulate the tree
    interp alias {} ::$name {} ::struct::tree::TreeProc $name

    return $name
}

##########################
# Private functions follow

# ::struct::tree::TreeProc --
#
#	Command that processes all tree object commands.
#
# Arguments:
#	name	Name of the tree object to manipulate.
#	cmd	Subcommand to invoke.
#	args	Arguments for subcommand.
#
# Results:
#	Varies based on command to perform

proc ::struct::tree::TreeProc {name {cmd ""} args} {
    # Do minimal args checks here
    if { $cmd == "" } {
	error "wrong # args: should be \"$name option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    if { [llength [info commands ::struct::tree::_$cmd]] == 0 } {
	variable commands
	set optlist [join $commands ", "]
	set optlist [linsert $optlist "end-1" "or"]
	error "bad option \"$cmd\": must be $optlist"
    }
    eval [list ::struct::tree::_$cmd $name] $args
}

# ::struct::tree::_append --
#
#	Appends a value to one of the keyed values associated with
#       n node. If no key is specified, the key data is assumed.
#
# Arguments:
#	name	   Name of the tree object.
#	node	   Node to look up.
#       -key key   key to append value to. Default to "data"
#       value      value to append
#
# Results:
#	The value of the key after appending to it.

proc ::struct::tree::_append {name node args} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [llength $args] != 1 && [llength $args] != 3 } {
	error "wrong # args: should be \"$name set $node ?-key key?\
		value\""
    }

    if { [llength $args] == 3 } {
	foreach {flag key} $args break
	if { ![string equal $flag "-key"] } {
	    error "invalid option \"$flag\": should be -key"
	}
    } else {
	set key "data"
    }

    if {[::Tree::attributeExist $theNode $key]} then {
      return [::Tree::appendAttribute $theNode $key [lindex $args end]]
    } else {
      error "invalid key \"$key\" for node \"$node\""
    }
}

# ::struct::tree::_children --
#
#	Return the child list for a given node of a tree.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#
# Results:
#	children	List of children for the node.

proc ::struct::tree::_children {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    return [::Tree::attribute $theNode _childList]
}

# ::struct::tree::_cut --
#
#	Destroys the specified node of a tree, but not its children.
#	These children are made into children of the parent of the
#	destroyed node at the index of the destroyed node.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up and cut.
#
# Results:
#	None.

proc ::struct::tree::_cut {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [::Tree::isRoot $theNode] } {
	# Can't delete the special root node
	error "cannot cut root node"
    }

    set insertChildren [::Tree::attribute $theNode _childList]

    set parent [::Tree::parent $theNode]
    set parentChildList [::Tree::attribute $parent _childList]
    set index [lsearch -exact $parentChildList $node]

    ::Tree::prune $theNode
    foreach child [::Tree::children $theNode] {
      ::Tree::graft $parent $child
    }
    set parentChildList [eval lreplace {$parentChildList} $index $index $insertChildren]
    ::Tree::attribute $parent _childList $parentChildList
    ::Tree::destroy $theNode
    unset -nocomplain nodes($node)
    return
}

# ::struct::tree::_delete --
#
#	Remove a node from a tree, including all of its values.  Recursively
#	removes the node's children.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to delete.
#
# Results:
#	None.

proc ::struct::tree::_delete {name node} {
    upvar 0 ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [::Tree::isRoot $theNode] } {
	# Can't delete the special root node
	error "cannot delete root node"
    }

    set theParent [::Tree::parent $theNode]
    set parentChildren [::Tree::attribute $theParent _childList]
    set oldIndex [lsearch $parentChildren $node]
    set parentChildren [lreplace $parentChildren $oldIndex $oldIndex]
    ::Tree::attribute $theParent _childList $parentChildren

    proc removeNodeVar { nodes tree depth } {
      upvar $nodes names2nodes
      unset names2nodes([Tree::name $tree])
      return 0
    }
    ::Tree::walk $theNode depth -postCall {removeNodeVar nodes}
    ::Tree::destroy $theNode
    return
}

# ::struct::tree::_depth --
#
#	Return the depth (distance from the root node) of a given node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to find.
#
# Results:
#	depth	Number of steps from node to the root node.

proc ::struct::tree::_depth {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    return [::Tree::depth $theNode]
}

# ::struct::tree::_destroy --
#
#	Destroy a tree, including its associated command and data storage.
#
# Arguments:
#	name	Name of the tree to destroy.
#
# Results:
#	None.

proc ::struct::tree::_destroy {name} {
    namespace delete ::struct::tree::tree$name
    interp alias {} ::$name {}
}

# ::struct::tree::_exists --
#
#	Test for existance of a given node in a tree.
#
# Arguments:
#	name	Name of the tree to query.
#	node	Node to look for.
#
# Results:
#	1 if the node exists, 0 else.

proc ::struct::tree::_exists {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      return 1
    } else {
      return 0
    }
}

# ::struct::tree::__generateUniqueNodeName --
#
#	Generate a unique node name for the given tree.
#
# Arguments:
#	name	Name of the tree to generate a unique node name for.
#
# Results:
#	node	Name of a node guaranteed to not exist in the tree.

proc ::struct::tree::__generateUniqueNodeName {name} {
    upvar ::struct::tree::tree${name}::nextUnusedNode nextUnusedNode
    while {[_exists $name "node${nextUnusedNode}"]} {
	incr nextUnusedNode
    }
    return "node${nextUnusedNode}"
}

# ::struct::tree::_get --
#
#	Get a keyed value from a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to query.
#	flag	Optional flag specifier; if present, must be "-key".
#	key	Optional key to lookup; defaults to data.
#
# Results:
#	value	Value associated with the key given.

proc ::struct::tree::_get {name node {flag -key} {key data}} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if {[::Tree::attributeExist $theNode $key]} then {
      return [::Tree::attribute $theNode $key]
    } else {
      error "invalid key \"$key\" for node \"$node\""
    }
}

# ::struct::tree::_getall --
#
#	Returns a serialized list of key/value pairs (suitable for use with [array set])
#       for the node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to query.
#
# Results:
#	value	serialized key/value pair listing.

proc ::struct::tree::_getall {name node args} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [llength $args] } {
	error "wrong # args: should be \"$name getall $node\""
    }
    return [::Tree::attribute $theNode]
}

# ::struct::tree::_index --
#
#	Determine the index of node with in its parent's list of children.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to look up.
#
# Results:
#	index	The index of the node in its parent

proc ::struct::tree::_index {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [::Tree::isRoot $theNode] } {
	# The special root node has no parent, thus no index in it either
	error "cannot determine index of root node"
    }
    return [lsearch -exact [::Tree::attribute [::Tree::parent $theNode] _childList] $node]
}

# ::struct::tree::_insert --
#
#	Add a node to a tree; if the node(s) specified already exist, they
#	will be moved to the given location.
#
# Arguments:
#	name		Name of the tree.
#	parentNode	Parent to add the node to.
#	index		Index at which to insert.
#	args		Node(s) to insert.  If none is given, the routine
#			will insert a single node with a unique name.
#
# Results:
#	nodes		List of nodes inserted.

proc ::struct::tree::_insert {name parentNode index args} {
    if { [llength $args] == 0 } {
	# No node name was given; generate a unique one
	set args [list [__generateUniqueNodeName $name]]
    } else {
	# Validate the node names
	foreach child $args {
	    if {[regexp "\[\r\t\n :\]" $child]} {
		return -code error "invalid node name \"$child\""
	    }
	}
    }

    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($parentNode)]} then {
      set parentTreeNode $nodes($parentNode)
    } else {
      error "parent node \"$parentNode\" does not exist in tree \"$name\""
    }

    # Make sure the index is numeric
    if { ![string is integer $index] } {
	# If the index is not numeric, set it the index where "end" would be
	set index [llength [::Tree::attribute $parentTreeNode _childList]]
    }

    set insertIndex $index
    foreach node $args {
	if {[info exists nodes($node)]} then {
	    # Move the node to its new home
	    if { [string equal $node "root"] } {
		error "cannot move root node"
	    }
            set theNode $nodes($node)

            # if the move is successful, then need to reindex siblings
            set theOldParent [::Tree::parent $theNode]

            if {[string equal [::Tree::graft $parentTreeNode $theNode] ""]} then {
              # move was unsuccessful
	      error "node \"$node\" cannot be its own descendant"
            } else {
              set OldParentChildren [::Tree::attribute $theOldParent _childList]
              set oldIndex [lsearch $OldParentChildren $node]
              set OldParentChildren [lreplace $OldParentChildren $oldIndex $oldIndex]
              ::Tree::attribute $theOldParent _childList $OldParentChildren
              
              # If the node is moving within its parent, and its old location
              # was before the new location, decrement the new location, so that
              # it gets put in the right spot
              if { [string equal $theOldParent $parentTreeNode] && ($oldIndex < $insertIndex) } {
                incr insertIndex -1
              }

              # insert into new parents child list
              set parentChildren [::Tree::attribute $parentTreeNode _childList]
              set parentChildren [linsert $parentChildren $insertIndex $node]
              ::Tree::attribute $parentTreeNode _childList $parentChildren

              incr insertIndex
            }
	} else {
	    # add the new node
            set nodes($node) [::Tree::graft $parentTreeNode [::Tree::create "$node"]]
            ::Tree::attribute $nodes($node) data ""
            # insert into new parents child list
            set parentChildren [::Tree::attribute $parentTreeNode _childList]
            set parentChildren [linsert $parentChildren $insertIndex $node]
            ::Tree::attribute $parentTreeNode _childList $parentChildren
            
            incr insertIndex
	}
    }
    return $args
}

# ::struct::tree::_isleaf --
#
#	Return whether the given node of a tree is a leaf or not.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#
# Results:
#	isleaf	True if the node is a leaf; false otherwise.

proc ::struct::tree::_isleaf {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    return [::Tree::isLeaf $theNode]
}

# ::struct::tree::_keys --
#
#	Returns a list of keys for the node.
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#
# Results:
#	value	A serialized list of key/value pairs.

proc ::struct::tree::_keys {name node args} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [llength $args] } {
	error "wrong # args: should be \"$name keys $node\""
    }
    array set dummy [::Tree::attribute $theNode]
    return [array names dummy]
}

# ::struct::tree::_keyexists --
#
#	Return true if the specified key exists for the node.
#       If no key is specified, the key data is assumed. 
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#       flag
#       key     The key to check for existance
#
# Results:
#	1 if the key exists, 0 else.

proc ::struct::tree::_keyexists {name node {flag -key} {key data}} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { ![string equal $flag "-key"] } {
	error "invalid option \"$flag\": should be -key"
    }
    return [::Tree::attributeExist $theNode $key]
}

# ::struct::tree::_lappend --
#
#	Appends a value (as a list) to one of the keyed
#       values associated with an node. If no key is specified,
#       the key data is assumed. 
#
# Arguments:
#	name	Name of the tree object.
#	node	Node to look up.
#       key     The key to check for existance
#       value   the value to append to the list
#
# Results:
#	val	Value associated with the given key of the given node

proc ::struct::tree::_lappend {name node args} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [llength $args] != 1 && [llength $args] != 3 } {
	error "wrong # args: should be \"$name set $node ?-key key?\
		value\""
    }
    
    if { [llength $args] == 3 } {
	foreach {flag key} $args break
	if { ![string equal $flag "-key"] } {
	    error "invalid option \"$flag\": should be -key"
	}
    } else {
	set key "data"
    }
    set value [lindex $args end]
    set working [::Tree::attribute $theNode $key]
    lappend working $value
    ::Tree::attribute $theNode $key $working
    return $working
}

# ::struct::tree::_move --
#
#	Move a node (and all its subnodes) from where ever it is to a new
#	location in the tree.
#
# Arguments:
#	name		Name of the tree
#	parentNode	Parent to add the node to.
#	index		Index at which to insert.
#	node		Node to move; the node must exist in the tree.
#	args		Additional nodes to move; these nodes must exist
#			in the tree.
#
# Results:
#	None.

proc ::struct::tree::_move {name parentNode index node args} {
    # really this is the same as insert, provided all of the
    # nodes exist

    upvar ::struct::tree::tree${name}::names2nodes nodes

    # verify that parent node exists
    if {![info exists nodes($parentNode)]} then {
	error "parent node \"$node\" does not exist in tree \"$name\""
    }

    # check the node
    if {![info exists nodes($node)]} then {
	error "node \"$node\" does not exist in tree \"$name\""
    }

    # check the args
    foreach nodeName $args {
      if {![info exists nodes($nodeName)]} then {
  	error "node \"$nodeName\" does not exist in tree \"$name\""
      }
    }
    eval ::struct::tree::_insert $name $parentNode $index $node $args

    return
}

# ::struct::tree::_next --
#
#	Return the right sibling for a given node of a tree.
#
# Arguments:
#	name		Name of the tree object.
#	node		Node to retrieve right sibling for.
#
# Results:
#	sibling		The right sibling for the node, or null if node was
#			the rightmost child of its parent.

proc ::struct::tree::_next {name node} {
    # The 'root' has no siblings.
    if { [string equal $node "root"] } {
	return {}
    }

    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }
    
    # Locate the parent and our place in its list of children.
    set allNames [::Tree::attribute [::Tree::parent $theNode] _childList]
    set index [lsearch -exact $allNames $node]
    incr index
    return [lindex $allNames $index]
}

# ::struct::tree::_numchildren --
#
#	Return the number of immediate children for a given node of a tree.
#
# Arguments:
#	name		Name of the tree object.
#	node		Node to look up.
#
# Results:
#	numchildren	Number of immediate children for the node.
 
proc ::struct::tree::_numchildren {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    return [llength [::Tree::attribute $theNode _childList]]
}

# ::struct::tree::_parent --
#
#	Return the name of the parent node of a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to look up.
#
# Results:
#	parent	Parent of node $node

proc ::struct::tree::_parent {name node} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if {![info exists nodes($node)]} then {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    set theNode $nodes($node)
    return [::Tree::name [::Tree::parent $theNode]]
}

# ::struct::tree::_previous --
#
#	Return the left sibling for a given node of a tree.
#
# Arguments:
#	name		Name of the tree object.
#	node		Node to look up.
#
# Results:
#	sibling		The left sibling for the node, or null if node was 
#			the leftmost child of its parent.

proc ::struct::tree::_previous {name node} {
    # The 'root' has no siblings.
    if { [string equal $node "root"] } {
	return {}
    }

    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }
    
    # Locate the parent and our place in its list of children.
    set allNames [::Tree::attribute [::Tree::parent $theNode] _childList]
    set index [lsearch -exact $allNames $node]
    incr index -1
    return [lindex $allNames $index]
}

# ::struct::tree::_set --
#
#	Set or get a value for a node in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to modify or query.
#	args	Optional arguments specifying a key and a value.  Format is
#			?-key key? ?value?
#		If no key is specified, the key "data" is used.
#
# Results:
#	val	Value associated with the given key of the given node

proc ::struct::tree::_set {name node args} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }

    if { [llength $args] > 3 } {
	error "wrong # args: should be \"$name set $node ?-key key?\
		?value?\""
    }
    
    set key "data"
    set haveValue 0
    if { [llength $args] > 1 } {
	foreach {flag key} $args break
	if { ![string match "${flag}*" "-key"] } {
	    error "invalid option \"$flag\": should be key"
	}
	if { [llength $args] == 3 } {
	    set haveValue 1
	    set value [lindex $args end]
	}
    } elseif { [llength $args] == 1 } {
	set haveValue 1
	set value [lindex $args end]
    }

    if { $haveValue } {
	# Setting a value
        ::Tree::attribute $theNode $key $value
	return $value
    } else {
	# Getting a value
        # FINK no error check
        if {[::Tree::attributeExist $theNode $key]} then {
          return [::Tree::attribute $theNode $key]
        } else {
          error "invalid key \"$key\" for node \"$node\""
        }
    }
}

# ::struct::tree::_size --
#
#	Return the number of descendants of a given node.  The default node
#	is the special root node.
#
# Arguments:
#	name	Name of the tree.
#	node	Optional node to start counting from (default is root).
#
# Results:
#	size	Number of descendants of the node.

proc ::struct::tree::_size {name {node root}} {
    upvar ::struct::tree::tree${name}::names2nodes nodes
    if {[info exists nodes($node)]} then {
      set theNode $nodes($node)
    } else {
      error "node \"$node\" does not exist in tree \"$name\""
    }
    
    # If the node is the root, we can do the cheap thing and just count the
    # number of nodes (excluding the root node) that we have in the tree with
    # array names
    if { [string equal $node "root"] } {
	set size [array size nodes]
	return [expr $size - 1]
    }

    # Otherwise we have to do it the hard way and do a full tree search
    set size 0
    proc countIt { varname tree depth } {
      upvar $varname it
      incr it
      return 0
    }

    ::Tree::walk $theNode level -postCall "countIt size"
    return [expr $size - 1]
}

# ::struct::tree::_splice --
#
#	Add a node to a tree, making a range of children from the given 
#	parent children of the new node.
#
# Arguments:
#	name		Name of the tree.
#	parentNode	Parent to add the node to.
#	from		Index at which to insert.
#	to		Optional end of the range of children to replace.
#			Defaults to 'end'.
#	node		Optional node name; if given, must be unique.  If not
#			given, a unique name will be generated.
#
# Results:
#	node		Name of the node added to the tree.

proc ::struct::tree::_splice {name parentNode from {to end} args} {
    upvar ::struct::tree::tree${name}::names2nodes nodes

    if { [llength $args] == 0 } {
	# No node name given; generate a unique node name
	set node [__generateUniqueNodeName $name]
    } else {
	set node [lindex $args 0]
    }

    set theParentNode $nodes($parentNode)

    if {[info exists nodes($node)]} then {
      error "node \"$node\" already exists in tree \"$name\""
    }

    # create the new node
    set newNode [::Tree::create $node]
    ::Tree::attribute $newNode data ""

    # move the nodes
    set parentChildList [::Tree::attribute $theParentNode _childList]
    set moveList [lrange $parentChildList $from $to]
    foreach child $moveList {
      ::Tree::graft $newNode [::Tree::children $theParentNode $child]
    }
    # add the new one
    ::Tree::graft $theParentNode $newNode

    # set the name order for the new child
    ::Tree::attribute $newNode _childList $moveList

    # remove the names from the parent
    set parentChildList [lreplace $parentChildList $from $to $node]
    ::Tree::attribute $theParentNode _childList $parentChildList

    set nodes($node) $newNode
    return $node
}

# ::struct::tree::_swap --
#
#	Swap two nodes in a tree.
#
# Arguments:
#	name	Name of the tree.
#	node1	First node to swap.
#	node2	Second node to swap.
#
# Results:
#	None.

proc ::struct::tree::_swap {name node1 node2} {
    # Can't swap the magic root node
    if { [string equal $node1 "root"] || [string equal $node2 "root"] } {
	error "cannot swap root node"
    }
    
    # Can only swap two real nodes
    if { ![_exists $name $node1] } {
	error "node \"$node1\" does not exist in tree \"$name\""
    }
    if { ![_exists $name $node2] } {
	error "node \"$node2\" does not exist in tree \"$name\""
    }

    # Can't swap a node with itself
    if { [string equal $node1 $node2] } {
	error "cannot swap node \"$node1\" with itself"
    }

    upvar ::struct::tree::tree${name}::names2nodes nodes

    set theNode1 $nodes($node1)
    set theNode2 $nodes($node2)

    set node1Parent [::Tree::parent $theNode1]
    set node2Parent [::Tree::parent $theNode2]

    # prune off the nodes
    ::Tree::prune $theNode1
    ::Tree::prune $theNode2

    set node1Children [::Tree::children $theNode1]
    set node1ChildList [::Tree::attribute $node1 _childList]
    set node2Children [::Tree::children $theNode2]
    set node2ChildList [::Tree::attribute $node2 _childList]

    # prune off all of the children
    foreach child [concat $node1Children $node2Children] {
      ::Tree::prune $child
    }

    # swap the nodes
    # -- first correct parent child list
    set parent1ChildList [::Tree::attribute $node1Parent _childList]
    set index1 [lsearch -exact $parent1ChildList $node1]

    set parent2ChildList [::Tree::attribute $node2Parent _childList]
    set index2 [lsearch -exact $parent2ChildList $node2]

    set parent1ChildList [lreplace $parent1ChildList $index1 $index1 $node2]
    ::Tree::attribute $node1Parent _childList $parent1ChildList

    # get this again in case the parents are the same
    set parent2ChildList [::Tree::attribute $node2Parent _childList]
    set parent2ChildList [lreplace $parent2ChildList $index2 $index2 $node1]
    ::Tree::attribute $node2Parent _childList $parent2ChildList

    if {$node1Parent == $theNode2} then {
      ::Tree::graft $theNode1 $theNode2
      set index1 [lsearch -exact $node2ChildList $node1]
      set node2ChildList [lreplace $node2ChildList $index1 $index1 $node2]
    } else {
      ::Tree::graft $node1Parent $theNode2
    }
    if {$node2Parent == $theNode1} then {
      ::Tree::graft $theNode2 $theNode1
      set index1 [lsearch -exact $node1ChildList $node2]
      set node1ChildList [lreplace $node1ChildList $index1 $index1 $node1]
    } else {
      ::Tree::graft $node2Parent $theNode1
    }

    # -- swap nodes corrected child list
    ::Tree::attribute $theNode1 _childList $node2ChildList
    ::Tree::attribute $theNode2 _childList $node1ChildList

    # make child of node 1 the children of node 2
    foreach child $node1Children {
      ::Tree::graft $theNode2 $child
    }
    # make child of node 2 the children of node 1
    foreach child $node2Children {
      ::Tree::graft $theNode1 $child
    }
    return
}

# ::struct::tree::_unset --
#
#	Remove a keyed value from a node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node to modify.
#	args	Optional additional args specifying which key to unset;
#		if given, must be of the form "-key key".  If not given,
#		the key "data" is unset.
#
# Results:
#	None.

proc ::struct::tree::_unset {name node {flag -key} {key data}} {
    if { ![_exists $name $node] } {
	error "node \"$node\" does not exist in tree \"$name\""
    }
    
    if { ![string match "${flag}*" "-key"] } {
	error "invalid option \"$flag\": should be \"$name unset\
		$node ?-key key?\""
    }

    upvar ::struct::tree::tree${name}::names2nodes nodes
    set theNode $nodes($node)
    if {[::Tree::attributeExist $theNode $key]} then {
      ::Tree::attributeUnset $theNode $key
      return 0
    } else {
      return 1
    }
    return 0
}

# ::struct::tree::_walk --
#
#	Walk a tree using a pre-order depth or breadth first
#	search. Pre-order DFS is the default.  At each node that is visited,
#	a command will be called with the name of the tree and the node.
#
# Arguments:
#	name	Name of the tree.
#	node	Node at which to start.
#	args	Optional additional arguments specifying the type and order of
#		the tree walk, and the command to execute at each node.
#		Format is
#		    ?-type {bfs|dfs}? ?-order {pre|post|in|both}? -command cmd
#
# Results:
#	None.

proc ::struct::tree::_walk {name node args} {
    set usage "$name walk $node ?-type {bfs|dfs}? ?-order {pre|post|in|both}? -command cmd"

    if {[llength $args] > 6 || [llength $args] < 2} {
	error "wrong # args: should be \"$usage\""
    }

    if { ![_exists $name $node] } {
	error "node \"$node\" does not exist in tree \"$name\""
    }

    # Set defaults
    set type dfs
    set order pre
    set cmd ""

    for {set i 0} {$i < [llength $args]} {incr i} {
	set flag [lindex $args $i]
	incr i
	if { $i >= [llength $args] } {
	    error "value for \"$flag\" missing: should be \"$usage\""
	}
	switch -glob -- $flag {
	    "-type" {
		set type [string tolower [lindex $args $i]]
	    }
	    "-order" {
		set order [string tolower [lindex $args $i]]
	    }
	    "-command" {
		set cmd [lindex $args $i]
	    }
	    default {
		error "unknown option \"$flag\": should be \"$usage\""
	    }
	}
    }
    
    # Make sure we have a command to run, otherwise what's the point?
    if { [string equal $cmd ""] } {
	error "no command specified: should be \"$usage\""
    }

    # Validate that the given type is good
    switch -glob -- $type {
	"dfs" {
	    set type "dfs"
	}
	"bfs" {
	    set type "bfs"
	}
	default {
	    error "invalid search type \"$type\": should be dfs, or bfs"
	}
    }
    
    # Validate that the given order is good
    switch -glob -- $order {
	"pre" {
	    set order pre
            if {$type == "dfs"} then {
              set CMD "-preCall localPreDFS"
            } else {
              set CMD "-preCall localPreBFS"
            }
	}
	"post" {
	    set order post
            if {$type == "dfs"} then {
              set CMD "-postCall localPostDFS"
            } else {
              set CMD "-postCall localPostBFS"
            }
	}
	"in" {
	    set order in
              set CMD "-inCall localIn"
 	}
	"both" {
	    set order both
            if {$type == "dfs"} then {
              set CMD "-preCall localPreDFS -postCall localPostDFS"
            } else {
              set CMD "-preCall localPreBFS -postCall localPostBFS"
            }
	}
	default {
	    error "invalid search order \"$order\":\
		    should be pre, post, both, or in"
	}
    }

    if {[string equal $order "in"] && [string equal $type "bfs"]} {
	error "unable to do a ${order}-order breadth first walk"
    }

    # Do the walk

    upvar ::struct::tree::tree${name}::names2nodes nodes

    # local routines to reformat the call to the users code
    proc localIn { tree depth index } {
      upvar name mytree
      upvar cmd command
      WalkCall $mytree [::Tree::name $tree] visit $command
      return 0
    }

    proc localPreDFS { tree depth } {
      upvar name mytree
      upvar cmd command
      WalkCall $mytree [::Tree::name $tree] enter $command
      return 0
    }

    proc localPostDFS { tree depth } {
      upvar name mytree
      upvar cmd command
      WalkCall $mytree [::Tree::name $tree] leave $command
      return 0
    }

    proc localPreBFS { tree depth } {
      upvar name mytree
      upvar cmd command
      WalkCall $mytree [::Tree::name $tree] enter $command
      return 0
    }

    set postBFSList ""
    proc localPostBFS { tree depth } {
      upvar postBFSList BFSList
      set BFSList [linsert $BFSList 0 $tree]
      return 0
    }

    # needed to make sure that the uplevel in Walk Call is correct
    proc WalkCallPassthrough {tree node action cmd} {
      WalkCall $tree $node $action $cmd
    }

    proc arrangeChildren { tree } {
      return [::Tree::attribute $tree _childList]
    }

    if { [string equal $type "dfs"] } {
	# Depth-first walk, several orders of visiting nodes
	# (pre, post, both, in)
        eval ::Tree::walk $nodes($node) depth $CMD -arrange arrangeChildren
    } else {
	# Breadth first walk (pre, post, both)
	# No in-order possible. Already captured.
        eval ::Tree::walk $nodes($node) level $CMD -arrange arrangeChildren
        if {[string equal $order "post"] || [string equal $order "both"]} then {
          foreach child $postBFSList {
            WalkCallPassthrough $name [::Tree::name $child] leave $cmd
          }
        }
    }
    return
}

# ::struct::tree::WalkCall --
#
#	Helper command to 'walk' handling the evaluation
#	of the user-specified command. Information about
#	the tree, node and current action are substituted
#	into the command before it evaluation.
#
# Arguments:
#	tree	Tree we are walking
#	node	Node we are at.
#	action	The current action.
#	cmd	The command to call, already partially substituted.
#
# Results:
#	None.

proc ::struct::tree::WalkCall {tree node action cmd} {
    uplevel 4 [string map [list \
	    %n [list $node]	\
	    %a [list $action]	\
	    %t [list $tree]	\
	    %% %] 		\
	    $cmd]
    return
}
