#
# Tree handling routines
#
# Author: Evan Rempel  <erempel@uvic.ca>
# http://web.uvic.ca/~erempel/tcl


package provide Tree 2.12

namespace eval ::Tree {

  namespace export isTree isRoot isLeaf
  namespace export create destroy root
  namespace export prune graft
  namespace export parent child children exist
  namespace export name value list appendValue
  namespace export height depth format walk

  variable Debug 0
}


# Object nodes have
#   _name       ell name
#   _var.name   Ell# ...  (indecies all children with same name)
#   _value.*    any value
#   _parent     Ell#
#   _children   Ell# Ell# ...

# ---------------------------------------------------------- token
# Create an Tree token
# TCL 8.4 allows for 64 bit integers. Prior to that
# a float is used to get more node names.
if {$tcl_version >= 8.4} {
  proc ::Tree::token { } {
      variable TreeObject
      if {![info exists TreeObject(tokenCount)]} {
  	set TreeObject(tokenCount) [expr wide(0)]
      }
      return [namespace current]::node[incr TreeObject(tokenCount)]
  }
} else {
  proc ::Tree::token { } {
      variable TreeObject
      if {![info exists TreeObject(tokenCount)]} {
  	set TreeObject(tokenCount) 0.0
      }
      set TreeObject(tokenCount) [expr {$TreeObject(tokenCount) + 1}]
      return [namespace current]::node[::format "%1.0f" $TreeObject(tokenCount)]
  }
}

# ---------------------------------------------------------- isTree
# Return True if treeToken represents a valid tree.
proc ::Tree::isTree { treeToken } {
  return [info exist $treeToken]
}

# ---------------------------------------------------------- isRoot
# Return True if treeToken has no parents.
proc ::Tree::isRoot { treeToken } {
  if {![info exist $treeToken]} { return 0 }
  if {[array names $treeToken _parent] == {}} { return 1 }
  return 0
}

# ---------------------------------------------------------- isLeaf
# isLeaf   token
# return a 1 if the tree is a leaf.
proc ::Tree::isLeaf { treeToken } {

  if {![info exist $treeToken]} { return 0 }
  if {[array names $treeToken _var.*] == {}} { return 1 }
  return 0
}


# ---------------------------------------------------------- create
# Create a single element tree
# Return the token for this tree
#
# ::Tree::create name ?value?
proc ::Tree::create { name args } {

    set token [::Tree::token]
    upvar 0 $token node

    set node(_name) $name
    if {$args != {}} then {
      set node(_value.value) [lindex $args 0]
    }
    return $token
}

# ---------------------------------------------------------- destroy
# Free up all of the resources for the tree and all of its
# children.
proc ::Tree::destroy { treeToken } {

    if {![::Tree::isRoot $treeToken]} then {
      ::Tree::prune $treeToken
    }
    ::Tree::internalDestroy $treeToken
    return ""
}

# ---------------------------------------------------------- destroy
# Destroy is recursive, but do not need to error check
# the children, only the passed in tree.
proc ::Tree::internalDestroy { tree } {
    upvar 0 $tree T
    if {[info exist T(_children)]} then {
      foreach child $T(_children) {
        ::Tree::internalDestroy $child
      }
    }
    unset -nocomplain $tree
}

# ---------------------------------------------------------- root
# return the root token for the tree containing token
# 
proc ::Tree::root { treeToken } {

  if {[info exist $treeToken]} then {
    set Root $treeToken
    upvar 0 $Root curNode
    while {[info exists curNode(_parent)]} {
      set Root $curNode(_parent)
      upvar 0 $Root curNode
    }
    return $Root
  } else {
    return ""
  }
}

# ---------------------------------------------------------- prune
# remove from the tree the entire branch starting at token
# making it a new tree.
proc ::Tree::prune { treeToken } {

  if {[info exist $treeToken] && ![::Tree::isRoot $treeToken]} then {
    upvar 0 $treeToken  node
    set oldParent $node(_parent)
    upvar 0 $oldParent oldParentNode

    # remove links from old parent
    set removeIndex [lsearch -exact -ascii $oldParentNode(_children) $treeToken]
    set oldParentNode(_children) [lreplace $oldParentNode(_children) $removeIndex $removeIndex]
    set name $node(_name)
    set removeIndex [lsearch -exact -ascii $oldParentNode(_var.$name) $treeToken]
    set oldParentNode(_var.$name) [lreplace $oldParentNode(_var.$name) $removeIndex $removeIndex]
    if {$oldParentNode(_var.$name) == ""} then {
      unset oldParentNode(_var.$name)
    }

    # remove links to parent
    unset -nocomplain node(_parent)
  }
  return $treeToken
}

# ---------------------------------------------------------- graft
# Attach child to the new parent.
# Return the token for this child object
# if an object of this name already exists, create a second node
# of the same name.
proc ::Tree::graft { parentToken childToken } {

  if {![::Tree::isTree $parentToken] || ![::Tree::isTree $childToken]} then {
    return ""
  }

  if {[::Tree::parent $childToken] == $parentToken} then {
    return $childToken
  }

  # ensure that not grafting an ancestor
  set Root $parentToken
  upvar 0 $Root curNode
  while {![string equal "$Root" "$childToken"] && [info exists curNode(_parent)]} {
    set Root $curNode(_parent)
    upvar 0 $Root curNode
  }
  if {[string equal "$Root" "$childToken"]} then {
    return ""
  }

  # -- not an ancestor, so do the prune/graft
  if {![::Tree::isRoot $childToken]} then {
    ::Tree::prune $childToken
  }

  upvar 0 $parentToken parentNode
  upvar 0 $childToken  childNode

  lappend parentNode(_children) $childToken
  set childNode(_parent) $parentToken

  lappend parentNode(_var.$childNode(_name)) $childToken

  return $childToken
}

# ---------------------------------------------------------- parent
# return the parent node for the object token
proc ::Tree::parent { treeToken } {

  upvar 0 $treeToken node
  if {[info exist node(_parent)]} then {
    return $node(_parent)
  } else {
    return ""
  }
}

# ---------------------------------------------------------- child
# Return the list of children of <b>treeToken</b> that are named
# name.
#
# O B S O L E T E
#
proc ::Tree::child { treeToken name } {

  upvar 0 $treeToken node
  if {[info exist node(_var.$name)]} then {
    return $node(_var.$name)
  } else {
    return ""
  }
}

# ---------------------------------------------------------- children
# return the list of children for the object token
proc ::Tree::children { treeToken args } {

  upvar 0 $treeToken node

  if {![info exist $treeToken]} then {
    return ""
  }

  if {[set max [llength $args]] == 0} {
    if {[info exist node(_children)]} then {
      return $node(_children)
    } else {
      return ""
    }
  } elseif {$max == 1} {
    set pattern $args
  } elseif {$max > 1} {
    set pattern "*"
    set sort ""
    set order ""
    set count 0
    while {$count < $max} {
      switch -exact -- [lindex $args $count] {
        "-pattern"
            { set pattern [lindex $args [incr count]]
              incr count
            }
        "-sort"
            { set sort [lindex $args [incr count]]
              incr count
            }
        "-order"
            { set order [lindex $args [incr count]]
              incr count
            }
        default
            { set pattern [lindex $args $count]
              incr count
            }
      }
    }

    set count 0
    if {$order != ""} { incr count }
    if {$sort != ""} { incr count }
    if {$count > 1} then {
      return -code error "only one of -order or -sort can be specified"
    }
    
    if {$sort != ""} then {
      set nameOrder ""
      foreach name [array names node _var.$pattern] {
        # chars 0..4 corespond to _var.
        lappend nameOrder [string range $name 5 end]
      }
      set nodeOrder ""
      foreach name [eval $sort {$nameOrder}] {
        if {[info exist node(_var.$name)]} then {
          set nodeOrder [concat $nodeOrder $node(_var.$name)]
        } else {
          lappend nodeOrder ""
        }
      }
      return $nodeOrder
    } elseif {$order != ""} then {
      # only look at those children that match the pattern
      array set matchedNames [array get node _var.$pattern]
      # create a list of children to return
      set nodeOrder ""
      foreach name $order {
        if {[info exist matchedNames(_var.$name)]} then {
          set nodeOrder [concat $nodeOrder $matchedNames(_var.$name)]
          # keep track that this has been added
          unset matchedNames(_var.$name)
        } else {
          # fill in an empty placeholder for the named node
          lappend nodeOrder ""
        }
      }
      # find nodes that are not yet included in the nodeOrder
      foreach {name value} [array get matchedNames] {
        set nodeOrder [concat $nodeOrder $value]
      }
      return $nodeOrder
    }
  }

  # -- just done to speed up retrieval of all children
  if {"$pattern" == "*"} then {
    if {[info exist node(_children)]} then {
      return $node(_children)
    } else {
      return ""
    }
  }
  set nodeOrder {}
  foreach {name value} [array get node "_var.$pattern"] {
    set nodeOrder [concat $nodeOrder $value]
  }
  return $nodeOrder
}

# ---------------------------------------------------------- exist
# exist   token  name ?childvar?
# return a 1 if the child exists. If childvar is included
# then set it to the list of token of the children matching name.
proc ::Tree::exist { treeToken name args} {

  if {[info exist $treeToken]} then {
    set retChild 0
    if {[llength $args] > 0} then {
      upvar [lindex $args 0] childNode
      set retChild 1
    }

    upvar 0 $treeToken node

    if {$retChild} then {
      set retValue 0
      set childNode ""
      foreach {child value} [array get node "_var.$name"] {
        set childNode [concat $childNode $value]
        set retValue 1
      }
    } else {
       if {[llength [array names node "_var.$name"]] > 0} then {
         set retValue 1
       } else {
         set retValue 0
       }
    }
    return $retValue
  }
  return 0
}

# ---------------------------------------------------------- name
# ::Tree::name Tree ?name?
# set and return the name of the node
proc ::Tree::name { treeToken args} {

  upvar 0 $treeToken node
  if {[info exist node]} then {
    if {[llength $args] > 0} then {
      set name [lindex $args 0]
      set oldName $node(_name)
      # if there is a parent, relink under new name
      if {[info exist node(_parent)]} then {
        upvar $node(_parent) parentNode
        lappend parentNode(_var.$name) $treeToken
        # remove named link from parent
        set removeIndex [lsearch -exact $parentNode(_var.$oldName) $treeToken]
        set parentNode(_var.$oldName) [lreplace $parentNode(_var.$oldName) $removeIndex $removeIndex]
      }
      set node(_name) $name
      return $oldName
    } elseif {[info exist node(_name)]} then {
      return $node(_name)
    } else {
      return ""
    }
  } else {
    return ""
  }
}

# ---------------------------------------------------------- attribute
# ::Tree::attribute tree ?attribute? ?value?
# set and return the value of the node
proc ::Tree::attribute { treeToken args} {

  upvar 0 $treeToken node
  if {[info exist node]} then {
    if {[llength $args] > 0} then {
      set attributeName [lindex $args 0]
      if {[llength $args] > 1} then {
        if {[info exist node(_value.$attributeName)]} then {
          set oldValue $node(_value.$attributeName)
        } else {
          set oldValue ""
        }
        set node(_value.$attributeName) [lindex $args 1]
        return $oldValue
      } elseif {[info exist node(_value.$attributeName)]} then {
        return $node(_value.$attributeName)
      } else {
        return ""
      }
    } else {
      set returnData ""
      foreach {key value} [array get node _value.*] {
        # character 0..6 corespond to _value.
        lappend returnData [string range $key 7 end] $value
      }
      return $returnData
    }
  } else {
    return ""
  }
}

# ---------------------------------------------------------- attributeUnset
# ::Tree::attributeUnset tree attribute
# set and return the value of the node
proc ::Tree::attributeUnset { treeToken attributeName } {

  upvar 0 $treeToken node
  set oldValue ""
  if {[info exist node]} then {
    if {[info exist node(_value.$attributeName)]} then {
      set oldValue $node(_value.$attributeName)
      unset node(_value.$attributeName)
    }
  }
  return $oldValue
}

# ---------------------------------------------------------- attributeExist
# ::Tree::attributeExist tree attribute
# set and return the value of the node
proc ::Tree::attributeExist { treeToken attributeName } {

  return [info exist ${treeToken}(_value.$attributeName)]
}

# ---------------------------------------------------------- list
# return the list of values pertaining to the multiple
# children called name of token
proc ::Tree::list { treeToken childName {attribute value}} {

  set RetList ""
  upvar 0 $treeToken node

  set returnList {}
  foreach {name children} [array get node "_var.$childName"] {
    foreach child $children {
      if {[info exist ${child}(_value.$attribute)]} then {
        upvar 0 $child tempNode
        lappend RetList $tempNode(_value.$attribute)
      }
    }
  }
  return $RetList
}

# ---------------------------------------------------------- appendAttribute 
# set the value of the node represented by token
proc ::Tree::appendAttribute { treeToken attribute value } {

  upvar 0 $treeToken node
  if {[info exist node]} then {
    return [append node(_value.$attribute) $value]
  } else {
    return ""
  }
}

# ---------------------------------------------------------- height
# return the number of levels in a tree
# a leaf tree has height of zero, one element, height of 1.
proc ::Tree::height { treeToken } {
  if {[info exist $treeToken]} then {
    return [::Tree::internalHeight $treeToken]
  } else {
    return ""
  }
}
# ---------------------------------------------------------- internalHeight
# no need to check all trees for existance
proc ::Tree::internalHeight { treeToken } {
  if {[::Tree::isLeaf $treeToken]} then {
    return 0
  } else {
    set max -1
    foreach child [::Tree::children $treeToken] {
      set height [::Tree::internalHeight $child]
      if {$height > $max} then {
        set max $height
      }
    }
    return [incr max]
  }
}

# ---------------------------------------------------------- depth
# return the depth that the treeToken is in the whole tree
proc ::Tree::depth { treeToken } {

  if {[info exist $treeToken]} then {
    set depth -1
    for {set Root $treeToken} {[info exist $Root]} { set Root [::Tree::parent $Root] } {
      incr depth
    }
    return $depth
  } else {
    return ""
  }
}

# ---------------------------------------------------------- format
# format  token
# format the xml object represented by token.
proc ::Tree::format { treeToken args} {
  variable output
  set output ""

  if {[llength $args] > 0} then {
    set valueName [lindex $args 0]
  } else {
    set valueName value
  }
  if {[info exist $treeToken]} then {
    return [::Tree::internalFormat $treeToken $valueName 0]
  } else {
    return ""
  }
}

# ---------------------------------------------------------- internalFormat
proc ::Tree::internalFormat { treeToken valueName level } {
  variable output

  set offsetString [string repeat " " $level]
  incr level

  upvar 0 $treeToken node
 
  set children [::Tree::children $treeToken]

  # create a string for all of the attributes other than $valueName
  array set attributes [::Tree::attribute $treeToken]
  set attribString ""
  foreach attributeName [array names attributes] {
    if {$attributeName != $valueName} then {
      append attribString " $attributeName=\"$attributes($attributeName)\""
    }
  }

  if {[info exist node(_value.$valueName)]} then {
    set value $node(_value.$valueName)
  } else {
    set value ""
  }

  if {[llength $children] == 0} then {
    if {$value == ""} then {
      append output "$offsetString$offsetString<$node(_name)$attribString/>\n"
    } else {
      append output "$offsetString$offsetString<$node(_name)$attribString>$value</$node(_name)>\n"
    }
  } else {
    append output "$offsetString$offsetString<$node(_name)$attribString>$value\n"
    foreach child $children {
      ::Tree::internalFormat $child $valueName $level
    }
    append output "$offsetString$offsetString</$node(_name)>\n"
  }
  return $output
}

# ---------------------------------------------------------- walk
# walk  token traversal -preCall command -inCall command -postCall command -order names

# walk through the tree following a {traversal} patern calling the commands 
#    traversal - depth or breadth
#    preCall - called pior to processing children
#    inCall  - called in between ALL children
#    postcall - called after all children 
# The children will be visted in the order of the -order argument

proc ::Tree::walk { token traversal args } {
  variable output
  set output ""

  if {[::Tree::isTree $token]} then {
    array set state {
        -preCall		""
        -inCall		""
        -postCall		""
        -order		""
        -arrange	""
        -traversal	""
        -singleInCall	1
        -sort		""
    }
    set state(-traversal) $traversal
  
    set options {-preCall -inCall -postCall -order -sort -arrange -singleInCall}
    set usage [join $options ", "]
    regsub -all -- - $options {} options
    set pat ^-([join $options "|"])$
    foreach {flag value} $args {
      if [regexp $pat $flag] {
        # Validate numbers
        if {[info exists state($flag)] && \
          [regexp {^[0-9]+$} $state($flag)] && \
          ![regexp {^[0-9]+$} $value]} {
          return -code error "Bad value for $flag ($value), must be integer"
        }
        set state($flag) $value
      } else {
        return -code error "Unknown option $flag, can be: $usage"
      }
    }
    set count 0
    if {$state(-order) != ""} { incr count }
    if {$state(-sort) != ""} { incr count }
    if {$state(-arrange) != ""} { incr count }
    if {$count > 1} then {
      return -code error "only one of -order, -sort or -order can be specified"
    }
    switch $traversal {
      depth { return [::Tree::internalDepthWalk $token state 0] }
      level { return [::Tree::internalLevelWalk $token state] }
    }
    return ""
  } else {
    return ""
  }
}

# ---------------------------------------------------------- internalDepthWalk
proc ::Tree::internalDepthWalk { tree stateVar depth } {
  upvar $stateVar state
  upvar $tree node

  if {[::Tree::isLeaf $tree]} then {

    # all three calls for leaf trees
    if {$state(-preCall) != ""} then {
      if {[uplevel [expr {$depth + 2}] $state(-preCall) $tree $depth]} then {
        return $tree
      }
    }
    if {$state(-inCall) != ""} then {
      if {[uplevel [expr {$depth + 2}] $state(-inCall) $tree $depth 0]} then {
        return $tree
      }
    }
    if {$state(-postCall) != ""} then {
      if {[uplevel [expr {$depth + 2}] $state(-postCall) $tree $depth]} then {
        return $tree
      }
    }
  } else {
    # -- not a leaf node
    set nodeOrder {}
    array set includeNodes {}

    if {$state(-sort) != ""} then {
      set nameOrder ""
      foreach name [array names node _var.*] {
        # chars 0..4 coresond to _var.
        lappend nameOrder [string range $name 5 end]
      }
      foreach name [eval $state(-sort) {$nameOrder}] {
        if {[info exist node(_var.$name)]} then {
          foreach child $node(_var.$name) {
            lappend nodeOrder $child
          }
        } else {
          lappend nodeOrder ""
        }
      }
    } elseif {($state(-order) != "") || ($state(-arrange) != "")} {
      # create a list of children to traverse
      if {$state(-arrange) != ""} {
        set startOrder [uplevel [expr {$depth + 2}] $state(-arrange) $tree]
      } else {
        set startOrder $state(-order)
      }
      foreach name $startOrder {
        if {[info exist node(_var.$name)]} then {
          foreach child $node(_var.$name) {
            lappend nodeOrder $child
            set includeNodes($child) 1
          }
        } else {
          # fill in an empty placeholder for the named node
          lappend nodeOrder ""
        }
      }
      # find nodes that are not yet included in the nodeOrder
      foreach child $node(_children) {
        if {![info exist includeNodes($child)]} then {
          lappend nodeOrder $child
        }
      }
    } else {
      set nodeOrder $node(_children)
    }

    set inFlag 0
    set nodeCount 0
    set inCallFlag 0
    set numberChildren [llength $nodeOrder]
    incr numberChildren -1

    foreach childNode $nodeOrder {

      # make the prefix call
      if {$nodeCount == 0} then {
        if {$state(-preCall) != ""} then {
          if {[uplevel [expr {$depth + 2}] $state(-preCall) $tree $depth]} then {
            return $tree
          }
        }
      }
      if {[::Tree::isTree $childNode]} then {
        set returnNode [::Tree::internalDepthWalk $childNode state [expr {$depth + 1}]]
        if {$returnNode != ""} then {
          return $returnNode
        }
      }

      # if prior to last child and multiple calls permitted, or this is the first call
      if {(($nodeCount < $numberChildren) && ($inCallFlag == 0)) || ($nodeCount == 0)} then {

        # in order call
        if {$state(-inCall) != ""} then {
          if {[uplevel [expr {$depth + 2}] $state(-inCall) $tree $depth $nodeCount]} then {
            return $tree
          }
        }
        set inCallFlag $state(-singleInCall)
      }
      if {$nodeCount == $numberChildren} then {

      # post order call
        if {$state(-postCall) != ""} then {
          if {[uplevel [expr {$depth + 2}] $state(-postCall) $tree $depth]} then {
            return $tree
          }
        }
      }
      incr nodeCount
    }
  }
  return ""
}

# ---------------------------------------------------------- internalLevelWalk
proc ::Tree::internalLevelWalk { tree stateVar } {
  upvar $stateVar state

  # -- queues for nodes to still visit
  #    use indirection to avoid copying large lists around
  set queue1 $tree
  set queue2 ""
  set workingQueue queue2
  set pendingQueue queue1
  set depth 0

  while {[eval llength \$$pendingQueue] > 0} {
    # move the pending queue into the working queue
    set tempQueue $workingQueue
    set workingQueue $pendingQueue
    set pendingQueue $tempQueue
    set $pendingQueue ""

    foreach tree [subst \$$workingQueue] {
      if {$state(-preCall) != ""} then {
        if {[uplevel 2 $state(-preCall) $tree $depth]} then {
          return $tree
        }
      }
      if {$state(-inCall) != ""} then {
        if {[uplevel 2 $state(-inCall) $tree $depth]} then {
          return $tree
        }
      }
      if {$state(-postCall) != ""} then {
        if {[uplevel 2 $state(-postCall) $tree $depth]} then {
          return $tree
        }
      }

      if {![::Tree::isLeaf $tree]} then {
        upvar 0 $tree node
 
        # create a list of children to traverse
        array set includeNodes {}
    
        if {$state(-sort) != ""} then {
          set nameOrder ""
          foreach name [array names node _var.*] {
            # chars 0..4 coresond to _var.
            lappend nameOrder [string range $name 5 end]
          }
          foreach name [eval $state(-sort) {$nameOrder}] {
            if {[info exist node(_var.$name)]} then {
              foreach child $node(_var.$name) {
                lappend $pendingQueue $child
              }
            } else {
              lappend nodeOrder ""
            }
          }
        } else {
          # create a list of children to traverse
          if {$state(-arrange) != ""} {
            set startOrder [uplevel 2 $state(-arrange) $tree]
          } elseif {$state(-order) != ""} {
            set startOrder $state(-order)
          } else {
            set startOrder $node(_children)
          }
          foreach name $startOrder {
            if {[info exist node(_var.$name)]} then {
              foreach child $node(_var.$name) {
                lappend $pendingQueue $child
                set includeNodes($child) 1
              }
            }
          }
          # find nodes that are not yet included in the nodeOrder
          foreach child $node(_children) {
            if {![info exist includeNodes($child)]} then {
              lappend $pendingQueue $child
            }
          }
        }
      }
    }
    incr depth
  }
  return ""
}